! Some Algorithms for the Conditional Mean and Covariance Matrix
!  John F Monahan, Dept of Statistics, NC State Univ, Sep 2005/Jul 2006
!
! Journal of Statistical Software 16(8), http://www.jstatsoft.org/v16/i08/
!
! Example 3 -- test program gcarma by comparing to two other methods
!   for constructing conditional mean vector and covariance matrix from
!   a subset of components from an ARMA process
!
!  1) construct joint covariance matrix and compute directly
!  2) use sweep operator -- easy to update changes in subset
!  3) faster method using qcarma (for ARMA covariance)
!
      program gsubex3
      implicit none
      integer, parameter :: n = 8       ! sample size
      integer, parameter :: s =  3      ! size of subset 
      integer, parameter :: tmodl = 2   ! number of test models 
      integer, parameter :: maxpq = 4   ! biggest p, q
      integer, dimension(n) :: list1,list2
      integer kmodl,p,q,i,j,k,m,mp1,idet  
      real, dimension(maxpq) :: phi,theta,yold,ynew
      real, dimension(maxpq+1) :: czz,bk
      real, dimension(n,n) :: A,A22
      real, dimension(s,s) :: A1d2,A1d2c2         ! A 1.2 and copy
      real, dimension(n,s) :: Ax,Ax2
      real, dimension(n-s) :: z2mmu2              ! z2 - mu2
      real, dimension(s) :: cmean                 ! A12*inv(A22)*z2mmu2
      real, dimension((s*(s+1))/2 ) :: R
      real, dimension(n) :: xcol
      real, dimension(2*n) :: avalues
      real ran,w,ww,det
!
  21  format(/' Test',i3,' with p=',i1,' and q=',i1)
  22  format(1x,i3,6f12.8)
  23  format(1x,5(' * *'))              ! just a border    
  24  format(' phi  ',6F11.6)                       
  25  format(' theta',6F11.6)                       
  26  format(8f10.6)
  29  format('max abs difference',f15.8)
  30  format('Stopped -- computationally singular matrix')
  31  format('random subsets -- S and T:',10i3/25i3)
  32  format(' original A matrix')
  33  format(' conditional covariance matrix A1.2 (direct)')
  34  format(' inv(A22)*A21  (direct)')
  35  format(' conditional covariance matrix A1.2 (sweep)')
  36  format(' inv(A22)*A21  (sweep)')
  37  format(' random conditional difference ( Z2 - mu2 )')
  38  format(' conditional mean'/' direct       qcarma')
  39  format(' inverse of conditional covariance matrix A1.2 (direct) ')
  40  format(' inverse of conditional covariance matrix A1.2 (qcarma) ')
!                     INTERFACE BLOCK
      INTERFACE
      SUBROUTINE CHLSKY(A,N,DET,IDET)
      REAL, DIMENSION(:,:), INTENT(IN OUT) :: A
      REAL, INTENT(OUT) :: DET
      INTEGER, INTENT(IN) :: N
      INTEGER, INTENT(OUT) :: IDET
      END SUBROUTINE CHLSKY
      SUBROUTINE CHLSHI(A,N,Y)
      REAL, DIMENSION(:,:), INTENT(IN) :: A
      INTEGER, INTENT(IN) :: N
      REAL, DIMENSION(:), INTENT(IN OUT) :: Y
      END SUBROUTINE CHLSHI
      SUBROUTINE CHLSIH(A,N,Y)                       
      REAL, DIMENSION(:,:), INTENT(IN) :: A
      REAL, DIMENSION(:), INTENT(IN OUT) :: Y
      INTEGER, INTENT(IN) :: N
      END SUBROUTINE CHLSIH
      SUBROUTINE CHLSOI(A,N)
      REAL, DIMENSION(:,:), INTENT(IN OUT) :: A
      INTEGER, INTENT(IN) :: N
      END SUBROUTINE CHLSOI
      subroutine sweepd(a,m,k,det,idet)
      real, dimension(:,:), intent(in out) :: a
      integer, intent(in) :: m,k
      integer, intent(in out) :: idet
      real, intent(in out) :: det
      end subroutine sweepd
      subroutine cfarma(phi,theta,p,q,bk,czz)
      implicit none
      integer, intent(in) :: p,q
      real, dimension(:), intent(in) :: phi, theta
      real, dimension(:), intent(out) :: czz,bk
      end subroutine cfarma
      subroutine qcarma(phi,theta,p,q,n,s,list,z2mmu2,r,cmean)
      implicit none
      integer, intent(in) :: p,q,n,s
      real, dimension(:), intent(in) :: phi, theta, z2mmu2
      real, dimension(:), intent(out) :: r,cmean
      integer, dimension(:), intent(in) :: list
      end subroutine qcarma
      END INTERFACE
!                     open file
      open( unit=6, file='gsubex3.out' )
!                   set seed to phone number
      w = ran(5151917)                                 
!                   generate a sample of models for testing
      do kmodl = 1, tmodl
!
!                   generate random p,q
      p = int( real(maxpq+1)*ran(kmodl) )  ! discrete unif( 0, maxpq)
      q = int( real(maxpq+1)*ran(kmodl) )  ! discrete unif( 0, maxpq)
      write(6,21) kmodl,p,q
!
!                   assign random parameters, insure stationary
!                   first the AR parameters
      if( p .ne. 0 ) then
      do k = 1,p
      ynew(k) = 2.*ran(k) - 1            ! uniform(-1,+1)
      if( k .gt. 1 ) then
        do i = 1,(k-1)
        ynew(i) = yold(i) + ynew(k)*yold(k-i)
        end do  ! loop on i
                     endif ! ( k .gt. 1 )
      yold = ynew
      end do ! loop on k
      phi = - yold
                     endif ! ( p .ne. 0 )
!                   second the MA parameters
      if( q .ne. 0 ) then
      do k = 1,q
      ynew(k) = 2.*ran(k) - 1            ! uniform(-1,+1)
      if( k .gt. 1 ) then
        do i = 1,(k-1)
        ynew(i) = yold(i) + ynew(k)*yold(k-i)
        end do  ! loop on i
                     endif ! ( k .gt. 1 )
      yold = ynew
      end do ! loop on k
      theta = - yold
                     endif ! ( q .ne. 0 )
!                   have ARMA parameters phi, theta -- write them
      if( p .ne. 0 ) write(6,24) (phi(j),j=1,p)
      if( q .ne. 0 ) write(6,25) (theta(j),j=1,q)
      m = max(p,q)
      mp1 = m + 1
!                             get covariance function
      czz = 0.                ! initialize
      call cfarma(phi,theta,p,q,bk,czz)         
!                             fill covariance matrix
!                              first values of A
      do i = 1,mp1
      avalues(i) = czz(i)
      end do ! loop on i
      do i = m+2,2*n
      avalues(i) = 0.
      if( p .ne. 0 ) then 
        do j = 1,p
        avalues(i) = avalues(i) + phi(j)*avalues(i-j)
        end do  ! on j
                     endif ! ( p .ne. 0 )
      end do  ! on i
!                             have values of A, now fill easily
      write(6,32)
      do i = 1,n
      do j = 1,n
      k = abs(i-j) + 1 
      a(i,j) = avalues(k)
      end do  ! on j
      write(6,26) (a(i,j),j=1,n)
      end do  ! on i
!
!                            generate random subset of size s
      j = s
      do i = 1,n
      if ( ran(i) .lt. real(j)/(n-i+1) ) then
         list1(s-j+1) = i
         j = j-1
                                       else   ! put in other list
         list2(i-s+j) = i
                                       end if ! ( ran .lt. j/(s-i+1) )
      end do  ! loop on i
!                             write out subsets
      write(6,31) (list1(j),j=1,s)
      write(6,31) (list2(j),j=1,n-s)
!
!                            FIRST METHOD -- BRUTE STRENGTH & IGNORANCE
!
!                            select subsets from A
      do i = 1,s
!                            first put A11 into A1d2
      do j = 1,s
      A1d2(i,j) = A( list1(i), list1(j) )
      end do  ! on j
!                            first put A21 into Ax
      do j = 1,n-s
      Ax(j,i) = A( list2(j), list1(i) )
      end do  ! on j
      end do  ! on i
!                            then load A22
      do i = 1,n-s
      do j = 1,n-s
      A22(i,j) = A( list2(i), list2(j) )
      end do  ! on j
      end do  ! on i
!                            compute Cholesky factor A22 = MM'
      call chlsky(A22,n-s,det,idet)
!                   check if computationally singular
      if( idet .lt. -2000000000 ) then
                                  write(6,30)
                                  stop
                                  end if ! ( idet .lt. -2000000000 )
!                            M-inverse * each col of A21
      do j = 1,s
      do i = 1,n-s
      xcol(i) = Ax(i,j)
      end do  ! on i
      call chlshi(A22,n-s,xcol)   ! solve equations
      do i = 1,n-s
      Ax(i,j) = xcol(i)
      end do  ! on i
      end do  ! on j
!                             Ax now holds M-inverse*A21
!   now get A1d2 = A11 - (M-inv*A21)'(M-inv*A21)
      write(6,33)
      do i = 1,s
      do j = 1,s
        do k = 1,n-s
        A1d2(i,j) = A1d2(i,j) - Ax(k,i)*Ax(k,j)
        end do  ! on k
      end do  ! on j
      write(6,26) (A1d2(i,j),j=1,s)
      end do  ! on i
      write(6,23)             ! write border
!   repeat task for inv(A22)*A21
!   Ax holds  M-inv-transpose * M-inverse * each col of A21
      do j = 1,s
      do i = 1,n-s
      xcol(i) = Ax(i,j)
      end do  ! on i
      call chlsih(A22,n-s,xcol)   ! solve equations
      do i = 1,n-s
      Ax(i,j) = xcol(i)
      end do  ! on i
      end do  ! on j
!                             write it out
      write(6,34)
      do i = 1,n-s
      write(6,26) (Ax(i,j),j=1,s)
      end do ! on i
      write(6,23)             ! write border
!                            
!                             SECOND METHOD -- USE SWEEP
!
      do i = 1,n-s
      k = list2(i)
      call sweepd(A,n,k,det,idet)
!                   check if computationally singular
      if( idet .lt. -2000000000 ) then
                                  write(6,30)
                                  stop
                                  end if ! ( idet .lt. -2000000000 )
      end do  ! on i
!                             COMPARE RESULTS
!
!                             write it out
!                             first A1d2 = Cov(x1 | x2)
      write(6,35)
      w = 0.
      do i = 1,s
      do j = 1,s
      A1d2c2(i,j) = A(list1(i),list1(j))
      w = max( w, abs( A1d2(i,j) - A1d2c2(i,j) ) )
      end do  ! on j
      write(6,26) (A1d2c2(i,j),j=1,s)
      do j = 1,n-s
      Ax2(j,i) = A(list2(j),list1(i))
      end do  ! on j
      end do  ! on i
!                             compare to previous
      write(6,29) w
      write(6,23)             ! write border
!                             now the corner inv(A22)*A21
      write(6,36)
      w = 0.
      do i = 1,n-s
      write(6,26) (Ax2(i,j),j=1,s)
      do j = 1,s
      w = max( w, abs( Ax(i,j) - Ax2(i,j) ) )
      end do  ! on j
      end do  ! on i
!                             compare to previous
      write(6,29) w
      write(6,23)             ! write border
!
!                             EXTRA BRUTE STRENGTH WORK
!                             get inverse of A1d2
      write(6,39)
      call chlsky(A1d2,s,det,idet)
      call chlsoi(A1d2,s)
      do i = 1,s
      write(6,26) (A1d2(i,j),j=1,s)
      end do  ! on i
!                             END of BRUTE STRENGTH EXTRAS
!
!                              something for the conditional diff
      do i = 1,n-s
      z2mmu2(i) = ran(i)  ! just list2 components
      end do  ! on i
      write(6,37)
      write(6,26) z2mmu2
!                             NOW USE GCARMA
!
!                             call gcarma to get R,cmean
      call qcarma(phi,theta,p,q,n,s,list1,z2mmu2,R,cmean)
!                             check covariance first
!                             inner product R'R to get inv(A1d2)
      write(6,40)
      w = 0.
      do i = 1,s
      do j = 1,s
        A(i,j) = 0.
        do k = 1,min(i,j)
        A(i,j) = A(i,j) + R((i*(i-1))/2 + k )*R( (j*(j-1))/2 + k) 
        end do  ! on k
      w = max( w, abs( A(i,j) - A1d2(i,j) ) )
      end do  ! on j
      write(6,26) (A(i,j),j=1,s)
      end do  ! on i
      write(6,29) w
!                             GCARMA'S CMEAN VS BRUTE STRENGTH METHOD
!
!                             last check conditional mean vector
!                             get A12*inv(A22)*z2mmu2
      write(6,38)
      w = 0.
      do i = 1,s
        ww = 0.
        do k = 1,n-s
        ww = ww + Ax(k,i)*z2mmu2(k)
        end do  ! on k
      w = max( w, abs( cmean(i)-ww ) )   ! note sign fixed
      write(6,26) ww,cmean(i)
      end do  ! on i
      write(6,29) w
!                             done with this parameter combo
      end do ! on kmodl
      stop
      end program gsubex3                  
      subroutine qcarma(phi,theta,p,q,n,s,list,z2mmu2,r,cmean)
!  construct conditional mean vector and covariance matrix
!  of subset given in LIST arising from an ARMA(p,q) time series
!
!  p     in integer      length of autoregression
!  q     in integer      length of moving average
!  phi   in real vector  autoregression parameters (length at least p)
!  theta in real vector  moving average parameters (length at least q)
!  n     in integer      length of time series
!  s     in integer      number in subset list 
!  list  in integ vector list of indices of subset (length at least s)
! z2mmu2 in real vector  difference of conditioning variables from mean
!  r     out real vector upper triang factor of A1.2 in sym storage
!                          R'R = inv(A1.2)
! cmean  out real vector conditional mean vector
!                          cmean = A12*inv(A22)*(z2-mu2)
!
!
! J F Monahan (Sept 2005) Dept of Statistics, NC State Univ
!
      implicit none
      integer, intent(in) :: p,q,n,s
      real, dimension(:), intent(in) :: phi, theta, z2mmu2
      real, dimension(:), intent(out) :: r,cmean
      integer, dimension(:), intent(in) :: list
      real, dimension(n,(s+1)) :: x             ! local matrix
      real t,rd
      integer i,j,k
!                     INTERFACE BLOCK
      INTERFACE
      subroutine qsarma(phi,theta,p,q,n,r,x)
      implicit none
      integer, intent(in) :: p,q,n,r
      real, dimension(:), intent(in) :: phi, theta
      real, dimension(:,:), intent(in out) :: x
      end subroutine qsarma
      END INTERFACE
!                             fill X matrix with elementary vectors
      X = 0         ! initialize to zero
      do i = 1,s
      X(list(i),i) = 1.
      end do  ! on i
!                             add zeros and z2-mu2 in last col
      k = 0
      do i = 1,n
      if( i .eq. list( min(s,k+1) ) ) then
        X(i,s+1) = 0.
        k = k + 1
                           else
        X(i,s+1) = z2mmu2(i-k)
                           end if ! ( i .eq. list(k) )
      end do  ! on i
!                             KEY STEP
!                             now apply half inverse of cov mx to X
      call qsarma(phi,theta,p,q,n,s+1,X)
!                        
!                             do QR via Householder
      do i = 1,s         
!                             start step i
      t = 0.
      do j = i,n
      t = t + X(j,i)*X(j,i)
      end do  ! on j
      t = sign(sqrt(t),X(i,i))          ! use X(i,i)'s sign
      R( (i*(i+1))/2 ) = -t   ! store on diagonal
      X(i,i) = X(i,i) + t
      rd = t*X(i,i)           ! reciprocal of d
!                             apply transf to X
      do j = i+1,s+1
!                             apply to col j
      t = 0.
      do k =  i,n
      t = t + X(k,i)*X(k,j)
      end do  ! on k
      t = t/rd
      do k = i,n
      X(k,j) = X(k,j) - t*X(k,i)
      end do  ! on k
!                             copy row entry
      if( j .le. s ) then
        R( (j*(j-1))/2 + i ) = X(i,j)   ! R is upper triangular
                     else
        cmean(i) = -X(i,j)    ! j = s+1, really u in cmean, sign!!
                     endif ! ( j .le. s )
!                             done with column j
      end do  ! on j
!                             done with column i
      end do  ! on i
!                             solve R*cmean = u
      do i = s,1,-1
      if ( i .ne. s ) then    ! is there a loop?
        do j = i+1,s
        cmean(i) = cmean(i) - R( (j*(j-1))/2 + i )*cmean(j)
        end do  ! on j
                      endif ! ( i .ne. s )
      cmean(i) = cmean(i) / R( (i*(i+1))/2 )
      end do  ! on i
      return
      end subroutine qcarma
      subroutine qsarma(phi,theta,p,q,n,r,x)
!  set-up for computing quadratic form in covariance matrix
!   of an ARMA(p,q) time series process
!
!  p     in integer      length of autoregression
!  q     in integer      length of moving average
!  phi   in real vector  autoregression parameters (length at least p)
!  theta in real vector  moving average parameters (length at least q)
!  n     in integer      number of rows of X (length of time series)
!  r     in integer      number of columns of X
!  X  *in*  real matrix  matrix to be operated on
!     *out*              L-inverse*X where Cov(Z) = L L'
! (L is Cholesky factor of covariance matrix of ARMA process)
!
      implicit none
      integer, intent(in) :: p,q,n,r
      real, dimension(:), intent(in) :: phi, theta
      real, dimension(:,:), intent(in out) :: x
      real, dimension(max(p,q)+1) :: czz,bk            ! local vector
      real, dimension(q+1) :: cma                  ! local vector
      real, dimension( (max(p,q)+1)*(max(p,q)+1) ) :: L  ! local matrix
      real t,s
      integer i,ip1,j,k,m,mp1,jll,pp1
!                     INTERFACE BLOCK
      INTERFACE
      subroutine cfarma(phi,theta,p,q,bk,czz)
      implicit none
      integer, intent(in) :: p,q
      real, dimension(:), intent(in) :: phi, theta
      real, dimension(:), intent(out) :: czz,bk
      end subroutine cfarma
      END INTERFACE
!                             one more
      m = max(p,q)
      mp1 = m + 1
      pp1 = p + 1
!                             first get ARMA covariance function
      call cfarma(phi,theta,p,q,bk,czz)
!                             get covariance function of MA part
      cma(1) = 1.
      if( q .ne. 0 ) then
        do i = 1,q
        cma(1) = cma(1) + theta(i)*theta(i)
        end do  ! loop on i
        do i = 1,q
        ip1 = i + 1 
        cma(ip1) = - theta(i)
        if( i .lt. q ) then
          do j = ip1,q
          cma(ip1) = cma(ip1) + theta(j)*theta(j-i)
          end do  ! loop on j
                       endif ! ( i .le. q )
         end do  ! loop on i
                     endif ! ( q .ne. 0 )
!                             done with preparations
!                             apply B to X (if needed -- p > 0 )
      if ( p .gt. 0 ) then 
!                             loop backwards from N to m+1
      do i = n,pp1,-1
        do j = 1,p            ! apply autoregressive operator
          do k = 1,r          !  to every column in X
          x(i,k) = x(i,k) - phi(j)*x(i-j,k)
          end do  ! loop on k
        end do  ! loop on j
      end do  ! loop on i
                      endif ! ( p .gt. 0 )
!  now the big step -- factor L on the fly, and solve L( ) = X
!   and overwrite X with solution
      do i = 1,n
      t = greta(i,i)          ! get (i,i) element of Cov(BZ)
      if ( i .gt. 1 ) then 
      jll = 1
      if ( i .gt. p ) jll = max(1,i-q)  ! lower limit for j
      if ( jll .le. i-1 ) then          ! is there a loop?
        do j = jll, i-1          
        s = greta(i,j)                  ! solving for new row of L
        if( j .gt. 1 ) then
          if( jll .le. j-1 ) then
          do k = jll,j-1
          s = s - L(loc(j,k))*L(loc(i,k))
          end do  ! on k
                             endif ! ( kll .le. j-1 )
                       endif ! ( jll .le. i-1 )
        L(loc(i,j)) = s / L(loc(j,j))             ! new row of L
        t = t - L(loc(i,j))*L(loc(i,j))
        end do  ! on j
                          endif ! ( jll .le. i-1 )
                      endif ! ( i .ne. 1 )
!                             ! get diagonal element of L
      L(loc(i,i)) = sqrt(t)
!                             ! now operate on BX
      if( i .gt. 1 ) then
      if ( jll .le. i-1 ) then          ! is there a loop?
        do j = jll, i-1
      do k = 1,r
        x(i,k) = x(i,k) - L(loc(i,j))*x(j,k)
      end do  ! on r
        end do  ! on j
                          endif ! ( jll .le. i-1 )
                      endif ! ( i .gt. 1 )
      do k = 1,r
      x(i,k) = x(i,k) / L(loc(i,i))
      end do  ! on r
      end do  ! loop on i ** main loop
      RETURN                                         
!                             local subprograms
      CONTAINS 
      integer function loc(i,j)    ! local function loc
!  storage location of Cholesky factor L
!  matrix is banded with max bandwith m+1 = mp1
!  three steps -- banded storage (i,j) in (i,j-i+m)
!              -- row mod m+1 since no need to store L
!              -- matrix stored as vector
      implicit none
      integer, intent(in) :: i,j
      loc = mp1*mod(i-1,mp1) + j - i + mp1
      return 
      end function loc             ! local function loc
      real function greta(i,j)     ! local function greta
      implicit none
      integer, intent(in) :: i,j
      integer k
! gives (i,j) element of covariance matrix of BZ
      k = abs(i-j) + 1
! most common call -- both i,j > max(p,q)
      greta = 0.
      if( k .le. q+1 ) greta = cma(k)
      if( ( i .gt. p ) .and. ( j .gt. p ) ) return
!                     other cases
      if( ( i .le. p ) .and. ( j .le. p ) ) then
!                     both i,j < m
        greta = czz(k)
                                            else
!                     remaining case
        greta = 0.
        if( k .le. q+1 ) greta = bk(k)
                                            endif ! both
      return
      end function greta           ! local function greta
!                             end of local subprograms
      end subroutine qsarma
      subroutine cfarma(phi,theta,p,q,bk,czz)
!   computes covariance function, sigma(0),...,sigma(m)   m=max(p,q)
! for ARMA process using McLeod's algorithm in Applied Statistics (1975)
!  Vol 24 No 2 pp255-256, Correction Vol 26 p194
!
! J F Monahan NCSU -- recoding in f95 (2004) of f77 version (1983)
!
!  p     in integer      length of autoregression
!  q     in integer      length of moving average
!  phi   in real vector  autoregression parameters (length at least p)
!  theta in real vector  moving average parameters (length at least q)
!  bk   out real vector  differenced covariance function
!                        bk(j+1) = Cov( Z(t), Z(t) - SUM phi(j)*Z(t-j) )
!  czz  out real vector  covariance function 
!                        czz(j+1) = Cov( Z(t), Z(t+j) ), j = 0,max(p,q)
!                        where Z follows ARMA(p,q) process
!
!     GAUSPP IS LINEAR EQUATIONS SOLVER              
      implicit none
      integer, intent(in) :: p,q
      real, dimension(:), intent(in) :: phi, theta
      real, dimension(:), intent(out) :: czz,bk
      real, dimension(max(p,q)+1,max(p,q)+1) :: a      ! local matrix
      real, dimension(max(p,q)+1) :: caz           ! local vector
      integer, dimension(max(p,q)+1) :: pi         ! local vector
      integer i,j,k,l,m,mp1
!                     INTERFACE BLOCK
      INTERFACE
      SUBROUTINE GAUSPP(A,N,PI)
      IMPLICIT NONE
      REAL, DIMENSION(:,:), INTENT(IN OUT) :: A
      INTEGER, INTENT (IN) :: N
      INTEGER, DIMENSION(:), INTENT(OUT) :: PI
      END SUBROUTINE GAUSPP
      SUBROUTINE GAUSSE(A,N,PI,B)
      IMPLICIT NONE
      REAL, DIMENSION(:,:), INTENT(IN) :: A
      REAL, DIMENSION(:), INTENT(IN OUT) :: B
      INTEGER, INTENT (IN) :: N
      INTEGER, DIMENSION(:), INTENT(IN) :: PI
      END SUBROUTINE GAUSSE
      END INTERFACE
!                             one more
      m = max(p,q)
      mp1 = m + 1
!                             first create vector of caz(k)
      caz(1) = 1.
      if( q .ne. 0 ) then
        do k = 1,q
        j = k+1
        caz(j) = - theta(k)
        if( p .ne. 0 ) then
          l = min(p,k)
          do i = 1, l
          caz(j) = caz(j) + phi(i)*caz(j-i)
          end do    ! loop on i
                       end if ! ( p .ne. 0 )
         end do  ! loop on k
                     end if ! ( q .ne. 0 )
!                             create vector b(k)
      bk = 0.
      bk(1) = caz(1)
      if( q .ne. 0 ) then
        do i = 1,q
        bk(1) = bk(1) - theta(i)*caz(i+1)
        end do  ! loop on i
        do k = 1,q
          j = k+1
          bk(j) = 0.
          do i = k,q
          bk(j) = bk(j) - theta(i)*caz(i-k+1)
          end do  ! loop on i
        end do  ! loop on k
                       end if ! ( q .ne. 0 )
!                             if p = 0 then finish
      czz = bk
      if ( p .ne. 0 ) then
!                             fill matrix A and rhs y=czz
      a = 0.                  ! initialize to zero
      do i = 1,mp1
      a(i,i) = 1.
        do j = 1,p
        l = abs(i-j-1)+1
        a(i,l) = a(i,l) - phi(j)
        end do  ! loop on j
      end do  ! loop on i
!                             this is negative of MacLeod's A
!                             now solve equations
!                             first factor
      call gauspp(a,m+1,pi)
!                             then solve
      call gausse(a,m+1,pi,czz)
                       end if ! ( p .ne. 0 )
      return
      end subroutine cfarma
      SUBROUTINE GAUSPP(A,N,PI)                                         
!  SUBROUTINE TO DO THE LU FACTORIZATION OF GAUSSIAN ELIMINATION
!  FACTORS P * A = L U WHERE P IS THE PERMUTATION MATRIX STORED IN PI
!  A IS DESTROYED AS L AND U ARE OVERWRITTEN ON IT
!  L IS UNIT LOWER AND U UPPER TRIANGULAR
!
!  USE GAUSSE TO SOLVE A SYSTEM OF LINEAR EQUATIONS
!
!    A      IN    MATRIX OF COEFFICIENTS N BY N
!           OUT   FACTORS L AND U OVERLAID
!    N      IN    TRUE SIZE OF MATRIX A (NOT DECLARED SIZE)
!           OUT   UNCHANGED
!    PI     IN    ANYTHING
!           OUT   PERMUTATION LIST OF N-1 PIVOT LOCATIONS
!                 PI(N) GIVES COMPUTED RANK, PI(N) = N IF NONSINGULAR
!
!  J F MONAHAN, SEPT 1985, N C S U, RALEIGH, NC 27695-8203 USA
!    REVISED JUNE 1999, APRIL 2000, FOR FORTRAN 95, JULY 2006
!
      IMPLICIT NONE
      REAL, DIMENSION(:,:), INTENT(IN OUT) :: A
      INTEGER, INTENT (IN) :: N
      INTEGER, DIMENSION(:), INTENT(OUT) :: PI
      REAL S,BIG                                               
      INTEGER I,J,K,KP1,NM1,IBIG
!
      PI(N) = 0
      IF( N .GT. 1 ) THEN
!                     DO NM1 STEPS TO FORM A TRIANGULAR SYSTEM
      NM1 = N - 1                                                       
!
      DO K = 1,NM1
!                     SEARCH FOR BEST PIVOT IN COLUMN K
      BIG = 0.                                                          
      IBIG = 0                                                          
      DO I = K,N
      S = ABS( A(I,K) )                                                 
      IF( S .GT. BIG ) THEN
!                     KEEP TRACK OF BEST I AND ITS SIZE
                       BIG = S
                       IBIG = I
                       ENDIF
      ENDDO         ! LOOP ON I
!                     TEST FOR SINGULARITY
!                             CHANGE 0. TO AN EPSILON IF DESIRED
      IF( BIG .EQ. 0. ) RETURN                                          
!                     KEEP TRACK OF THE RANK IN PI(N)
      PI(N) = K                                                         
!                     STORE PIVOT LOCATION IN PERMUTATION LIST
      PI(K) = IBIG                                                      
!                     SWITCH ONLY IF NECESSARY
      IF( IBIG .NE. K ) THEN
          DO J = 1,N
          S = A(K,J)
          A(K,J) = A(IBIG,J)
          A(IBIG,J) = S
          ENDDO     ! LOOP ON J
!
                       ENDIF
!                     ELIMINATE REMAINDER OF COLUMN K AND CREATE ONE
      KP1 = K + 1                                                       
      DO I = KP1,N
      A(I,K) = A(I,K) / A(K,K)                                          
          DO J = KP1,N
          A(I,J) = A(I,J) - A(I,K) * A(K,J)
          ENDDO     ! LOOP ON J
      ENDDO         ! LOOP ON I
         do i = 1,n
         enddo      ! loop on i
      ENDDO         ! END OF MAIN LOOP ON K
      ENDIF
!                     TEST LAST (AND MAYBE ONLY) ONE FOR SINGULARITY
      IF ( ABS( A(N,N) ) .GT. 0. ) THEN
!                     ONLY (COMPUTATIONALLY) NONSINGULAR MATRICES
!                     REACH HERE
          PI(N) = N
          ENDIF
      RETURN                                                            
      END SUBROUTINE GAUSPP
      SUBROUTINE GAUSSE(A,N,PI,B)
!
!   SUBROUTINE TO SOLVE A SYSTEM OF LINEAR EQUATIONS A * X = B
!
!     A     IN   HOLDS FACTORS L AND U OVERLAID FROM GAUSPP
!           OUT  UNCHANGED
!     N     IN   TRUE SIZE OF MATRIX A (NOT DECLARED SIZE)
!           OUT  UNCHANGED
!     PI    IN   LIST OF PIVOTS -- HOLDS PERMUTATION MATRIX P
!                IN FACTORED FORM
!           OUT  UNCHANGED
!
!  ***  GAUSPP MUST HAVE BEEN CALLED PREVIOUSLY  ***
!
!  GAUSPP FACTORS P * A = L U
!  TO SOLVE A X = B, FIRST PERMUTE B TO GET P A X = L U X = P B
!  THEN SOLVE TWO TRIANGULAR SYSTEMS OF EQUATIONS
!
!  J F MONAHAN, SEPT 1985, N C S U, RALEIGH, NC 27695 - 8203 USA
!     REVISED JUNE, 1999 FOR FORTRAN 95, JULY 2006
!
!                     DECLARATIONS FIRST
      IMPLICIT NONE
      REAL, DIMENSION(:,:), INTENT(IN) :: A
      REAL, DIMENSION(:), INTENT(IN OUT) :: B
      INTEGER, INTENT (IN) :: N
      INTEGER, DIMENSION(:), INTENT(IN) :: PI
      REAL S
      INTEGER I,J,K,II,IP1,IM1,NM1
!
      IF( N .GT. 1 ) THEN
      NM1 = N - 1                                                       
!                           PERMUTE B FIRST
      DO K = 1,NM1
      S = B(K)                                                          
      I = PI(K)                                                         
      B(K) = B(I)                                                       
      B(I) = S                                                          
      ENDDO         ! LOOP ON K
!                     L IS UNIT TRIANGULAR MATRIX
!                           SOLVE L * Y = P * B  FROM THE TOP DOWN
      DO I = 2,N
      IM1 = I - 1                                                       
          DO J = 1,IM1
          B(I) = B(I) - A(I,J) * B(J)
          ENDDO     ! LOOP ON J
      ENDDO         ! LOOP ON I
!                      U IS UPPER TRIANGULAR
!                           SOLVE U * X = Y
      ENDIF
      B(N) = B(N) / A(N,N)
!                        LAST ONE DONE FIRST
      IF( N .EQ. 1 ) RETURN                                             
!                        NOW FROM THE BOTTOM UP, I = N-1, ... , 2, 1
      DO II = 2,N
      I = N + 1 - II                                                    
      IP1 = I + 1                                                       
          DO J = IP1,N
          B(I) = B(I) - A(I,J) * B(J)
          ENDDO     ! LOOP ON J
      B(I) = B(I) / A(I,I)                                              
      ENDDO         ! LOOP ON II, I
      RETURN                                                            
      END SUBROUTINE GAUSSE
      REAL FUNCTION RAN(IDUM)
!  PORTABLE IMPLEMENTATION OF UNIFORM PSEUDORANDOM NUMBER GENERATOR
!    LEWIS, GOODMAN, & MILLER MULTIPLICATIVE GENERATOR
!     X(N+1) = MOD( 16807*X(N), 2**31-1 )
!
!  P. BRANTLEY, B.L. FOX, L. SCHRAGE (1983) A GUIDE TO SIMULATION
!   SPRINGER-VERLAG, NEW YORK.  PP. 200-202.
!  UPDATED VERSION OF
!  LINUS SCHRAGE (1979)'A MORE PORTABLE FORTRAN RANDOM NUMBER GENERATOR' 
!   ACM TRANSACTIONS ON MATHEMATICAL SOFTWARE, VOLUME 5, PP. 132-138.
!
!  ARGUMENT 
!   IDUM    INTEGER  FIRST CALL SETS SEED, IGNORED IN SUBSEQUENT CALLS
!
      IMPLICIT NONE
      INTEGER, INTENT(IN) :: IDUM
      REAL, PARAMETER :: RP = 4.656612875E-10     ! 1/P
      INTEGER, PARAMETER :: A = 16807             ! MULTIPLIER
      INTEGER, PARAMETER :: B = 127773            ! B = P / A
      INTEGER, PARAMETER :: C = 2836              ! C = P MOD A
      INTEGER, PARAMETER :: P = 2147483647        ! MODULUS 2**31 - 1
      INTEGER, SAVE :: IX = 0
      INTEGER K1
!
!                   IF NOT FIRST CALL, THEN SKIP SETTING SEED
      IF( IX .EQ. 0) IX = IDUM
!                   WRITE NUMBER AS ALPHA*2**16 + BETA
      K1 = IX / B
      IX = A*( IX - K1*B) - K1*C
!                   ABOVE DOES A*IX MOD B -K1*C
      IF( IX .LT. 0 ) IX = IX + P 
!                   RP BELOW IS RECIPROCAL OF P
      RAN = REAL(IX)*RP
      RETURN
      END FUNCTION RAN
      SUBROUTINE ADJUST(D,I)                                            
!  NORMALIZES D WHILE KEEPING CONSTANT VALUE OF D * (2**I)
      INTEGER, PARAMETER :: IBIG = -2147483644
      REAL, INTENT (IN OUT) :: D
      INTEGER, INTENT(IN OUT) :: I
      IF ( D .GT. 0.0 ) THEN
        DO WHILE ( D .LT. 1.0 )
        D = D*16.
        I = I-4
        ENDDO
        DO WHILE ( D .GT. 16.0 )
        D = D/16.
        I = I+4
        ENDDO
                        ELSE
      I = IBIG      ! IF D < 0 THEN I = - BIG
                        ENDIF
      RETURN
      END SUBROUTINE ADJUST
      SUBROUTINE CHLSKY(A,N,DET,IDET)                                   
!  CHLSKY COMPUTES THE CHOLESKY (SQUARE-ROOT)  FACTORIZATION
!   A = L * ( L - TRANSPOSE )   WHERE L IS LOWER TRIANGULAR
!     A IS OVERWRITTEN BY L IN ITS LOWER TRIANGULAR PART
!  SUBROUTINE ADJUST KEEPS DETERMINANT FROM EXPLODING USING
!  1 LE DET LE 16   AND   DETERMINANT OF A IS DET*2**IDET
!
!  J F MONAHAN (DEC,1983) DEPT OF STAT, N C S U, RALEIGH, N C 27650 USA
!      REVISED JANUARY, 2000 FOR FORTRAN 95
      IMPLICIT NONE
      REAL, DIMENSION(:,:), INTENT(IN OUT) :: A
      REAL, INTENT(OUT) :: DET
      INTEGER, INTENT(IN) :: N
      INTEGER, INTENT(OUT) :: IDET
      REAL T,S
      INTEGER I,J,K,IM1,JM1
!                                    INITIALIZE FOR DETERMINANT
      DET = 1.                                                        
      IDET = 0                                                          
!                                      DO I-TH ROW
      DO I = 1,N
      T = A(I,I)                                                        
!                                      FIRST ROW IS TRIVIAL
      IF(I .GT. 1) THEN
        IM1 = I-1
        DO J=1,IM1
        A(J,I) = 0.
!                                      WORK ON (I,J)-TH ELEMENT
        S = A(I,J)
        IF(J .GT. 1) THEN
          JM1 = J-1
          DO K = 1,JM1
          S = S - A(J,K) * A(I,K)
          ENDDO     ! LOOP ON K
                     ENDIF
        A(I,J)= S / A(J,J)
!                                 WORK ON DIAGONAL ELEMENT
        T = T - A(I,J) * A(I,J)
        ENDDO       ! LOOP ON J
!                                 FINISHED WITH LOOP ON J
!                                 UPDATE DETERMINANT WITH DIAGONAL
                   ENDIF
      DET = DET * T
!                                 KEEP DET FROM EXPLODING
      CALL ADJUST(DET,IDET)                                             
!                                 ABORT IF DET IS NEGATIVE OR ZERO
      IF(IDET.LT.-2000000000) RETURN                                    
!                                 FINISH A POSITIVE DIAGONAL
      A(I,I) = SQRT(T)                                                  
      ENDDO         ! LOOP ON I
      RETURN                                                            
      END SUBROUTINE CHLSKY
      SUBROUTINE CHLSHI(A,N,Y)                                          
!
!   OVERWRITES   Y   WITH      INVERSE( L ) * Y
!    WHERE L IS A LOWER TRIANGULAR MATRIX (N BY N) STORED IN A
!
!   TO COMPUTE QUADRATIC FORM IN THE INVERSE OF A POSITIVE DEFINITE
!   MATRIX A, THAT IS   Y' INVERSE( A ) Y THEN
!     FIRST:  FACTOR A BY CHLSKY(A,N,DET,IDET)
!             THIS COMPUTES LOWER TRIANGULAR L SO THAT A = L * L'
!     SECOND:  COMPUTE INVERSE(L) * Y BY CHLSHI(A,N,Y)
!     THIRD:  COMPUTE THE SUM OF SQUARES OF THE ELEMENTS OF Y
!
!  J F MONAHAN (JULY,1984) DEPT OF STAT, N C S U, RALEIGH, N C 27695 USA
!      REVISED JUNE, 1999 FOR FORTRAN 95
      IMPLICIT NONE
      REAL, DIMENSION(:,:), INTENT(IN) :: A
      REAL, DIMENSION(:), INTENT(IN OUT) :: Y
      INTEGER, INTENT(IN) :: N
      INTEGER I,J,IM1
!
!  SINCE THE MATRIX IS LOWER TRIANGULAR, SOLVE FROM THE TOP DOWN
!
      DO I = 1,N
!                            BRANCH AROUND ON THE FIRST ONE
      IF( I .GT. 1 ) THEN
        IM1 = I - 1
        DO J = 1,IM1
        Y(I) = Y(I) - A(I,J)*Y(J)
        ENDDO
                     ENDIF
      Y(I) = Y(I) / A(I,I)
      ENDDO         ! LOOP ON I
      RETURN                                                            
      END SUBROUTINE CHLSHI
      SUBROUTINE CHLSIH(A,N,Y)                                          
!                              ***  NOTICE TRANSPOSE  ***
!   OVERWRITES   Y   WITH      INVERSE( L' ) * Y
!    WHERE L IS A LOWER TRIANGULAR MATRIX (N BY N) STORED IN A
!
!   TO COMPUTE THE PRODUCT OF THE INVERSE OF A POSITIVE DEFINITE MATRIX
!   AND A VECTOR Y, THAT IS  INVERSE( A ) * Y  THEN
!     FIRST:  FACTOR A BY CHLSKY(A,N,DET,IDET)
!             THIS COMPUTES LOWER TRIANGULAR L SO THAT A = L * L'
!     SECOND:  COMPUTE INVERSE(L) * Y BY CHLSHI(A,N,Y)
!     THIRD:  COMPUTE INVERSE(L') * ( INVERSE(L)*Y ) BY CHLSIH(A,N,Y)
!
!  J F MONAHAN (JULY,1984) DEPT OF STAT, N C S U, RALEIGH, N C 27695 USA
!      REVISED JUNE, 1999 FOR FORTRAN 95
      IMPLICIT NONE
      REAL, DIMENSION(:,:), INTENT(IN) :: A
      REAL, DIMENSION(:), INTENT(IN OUT) :: Y
      INTEGER, INTENT(IN) :: N
      INTEGER I,II,IP1,J
!
!  SINCE THE MATRIX IS UPPER TRIANGULAR, SOLVE FROM THE BOTTOM UP
!
!                      DO THE LAST ONE SEPARATELY
      Y(N) = Y(N) / A(N,N)                                              
!                      IF N = 1 THEN QUIT ELSE START LOOPING
      IF( N .EQ. 1 ) RETURN                                             
      DO II = 2,N
!                      I BELOW NOW GOES FROM N-1 DOWN TO 1
        I = N + 1 - II
        IP1 = I + 1
        DO J = IP1,N
        Y(I) = Y(I) - A(J,I)*Y(J)
        ENDDO       ! LOOP ON J
        Y(I) = Y(I) / A(I,I)
      ENDDO         ! LOOP ON II,I
      RETURN                                                            
      END SUBROUTINE CHLSIH
      subroutine sweepd(a,m,k,det,idet)
! sweep operator on the sum of squares and cross products mx A
!  a  real matrix m by m  (dummy dimensioned -- use INTERFACE block)
!  m  integer size of matrix p+1 for one dependent and p regressors
!  k  integer variable to be swept
!
! update of SWEEP to include determinant calculation
!  use of ADJUST presupposes positive definite matrix
      implicit none
      real, dimension(:,:), intent(in out) :: a
      integer, intent(in) :: m,k
      integer, intent(in out) :: idet
      real, intent(in out) :: det
      real b,d
      integer i,j
!                   sweeping in/out variable k
      d = a(k,k)
      det = det * d
      call adjust(det,idet)
      do j = 1,m
      a(k,j) = a(k,j)/d
      end do  ! loop on j
!                   fixed row k, now rest of matrix
      do i = 1,m
      if( i .ne. k ) then
      b = a(i,k) 
      do j = 1,m
      a(i,j) = a(i,j) - b*a(k,j)
      end do  ! loop on j
      a(i,k) = -b/d
                     END if
      end do  ! loop on i
      a(k,k) = 1.d0/d 
!                   done
      return
      end subroutine sweepd
      SUBROUTINE CHLSOI(A,N)
!
!   OVERWRITES L WITH INVERSE(L')*INVERSE(L) WHERE L IS A LOWER
!   TRIAGULAR (N BY N) MATRIX STORED IN A
!
!   TO COMPUTE THE INVERSE OF A POSITIVE DEFINITE MATRIX A
!     FIRST:  FACTOR A = L * L' WHERE L IS LOWER TRIANGULAR
!             USING CHLSKY(A,N,D,IDET)
!     SECOND: COMPUTE THE INVERSE(A) = INVERSE(L') * INVERSE(L)
!             USING CHLSOI(A,N)
!
! J F MONAHAN (JULY,1984) DEPT OF STATISTICS, N C S U, RALEIGH, N C 2769
!     REVISED JANUARY, 2000 FOR FORTRAN 95
      REAL, DIMENSION(:,:), INTENT(IN OUT) :: A
      INTEGER, INTENT(IN) :: N
      REAL S
      INTEGER I,J,K,IM1,KP1
!                     FIRST INVERT L IN PLACE
      DO K = 1,N
!                     SOLVE L * X = E(K)
      A(K,K) = 1./A(K,K)                                                
!                     BRANCH AROUND IF K EQUALS N
      IF( K .LT. N ) THEN
        KP1 = K + 1
        DO I = KP1,N
        S = 0.
        IM1 = I - 1
          DO J = K,IM1
          S = S - A(I,J)*A(J,K)
          ENDDO ! LOOP ON J
        A(I,K) = S/A(I,I)
        ENDDO ! LOOP ON I
                      ENDIF
      ENDDO         ! LOOP ON K
!                    DONE WITH INVERSION, NOW COMPUTE INNER PRODUCT
      DO J = 1,N
        DO I = J,N
!                    GET (I,J) ELEMENT OF MATRIX IS THE INNER PRODUCT
!                    OF COLUMN I AND COLUMN J
        S = 0.
          DO K = I,N
          S = S + A(K,I)*A(K,J)
          ENDDO     ! LOOP ON K
!                    STORE THE INNER PRODUCT
        A(I,J) = S
        A(J,I) = S
        ENDDO       ! LOOP ON I
      ENDDO         ! LOOP ON J
      RETURN
      END SUBROUTINE CHLSOI


